//Board.java

/***
 Questa classe incapsula il tabellone di gioco, ad un livello astratto, provvedendo a fornire le funzioni per 
 gestire la modifica del suo stato. Board puu' essere anche utilizzata per salvare lo stato della partita in un 
 determinato istante.
 
 Autore: Alessandro Caviola
***/

import java.util.*;

class Board
{
  protected boolean next_move; /***Se true, allora e' il turno del giocatore 1, altrimenti dell'avversario***/
  protected short turn; /***Indica il numero di mosse fatte***/
  
  static public byte MASK_MAX = 3; // => 00000011 maschera per il valore limite della casella
  static public byte MASK_CUR = 28; // => 00011100 maschera per il valore corrente della casella
  static public byte MASK_OWN = 32; // => 00100000 maschera per il proprietario della casella
  
  protected byte[][] board = new byte[6][5];
  
  /***Costruisce un oggetto della classe Board.***/
  Board()
  {
    for(int x = 0; x <= 5; x++)
      for(int y = 0; y <= 4; y++)
    {
      if(x == 0 || x == 5)
      {
        if(y == 0 || y == 4) board[x][y] = Pack(1, 0, false);
        else board[x][y] = Pack(2, 0, false);
      } 
      else if(y == 0 || y == 4) board[x][y] = Pack(2, 0, false);
      else board[x][y] = Pack(3, 0, false);
    }
    
    next_move = false;
    turn = 0;
  }
  
  /***copia l'oggetto src nell'oggetto corrente.***/
  Board(Board src)
  {
    for(int x = 0; x <= 5; x++)
      for(int y = 0; y <= 4; y++)
    {
      board[x][y] = src.board[x][y];
    }
    
    next_move = src.next_move;
    turn = src.turn;
  }
  
  /***Copia lo stato di un oggetto Board nell'oggetto corrente.***/
  public boolean CopyFrom(Board src)
  {
    if(src == null) return false;
    next_move = src.next_move;
    turn = src.turn;
    for(int x = 0; x <= 5; x++)
      for(int y = 0; y <= 4; y++)
    {
      board[x][y] = src.board[x][y];
    }
    
    return true;
  }
  
  /***Restituisce una copia dell'oggetto corrente.***/
  public Board GetCopy()
  {
    return new Board(this);
  }
  
  public short GetTurn()
  {
    return turn;
  }
  
  public void SetTurn(int turn)
  {
    turn = (short) turn;
  }
  
  public boolean GetNextMove()
  {
    return next_move;
  }
  
  public void SetNextMove(boolean player)
  {
    next_move = player;
  }
  
  /***Compatta in un singolo byte le informazioni passate come paranmetri. La codifica avviene nel seguente modo: 
  i due bytes meno significativi codificano il primo parametro, i successivi tre bytes codificano il secondo parametro e il 
  sesto byte codifica l'ultimo parametro. Il metodo non verifica totalmente la validita' dei valori passati.***/
  public static byte Pack(byte limit, byte current, boolean player)
  {
    if(current > 4) current = 4;
  
    byte res = (byte)(limit + (current << 2));
    if(player) return (byte)(res + (1 << 5)); //player == true
    return res; //player == false; 
  }
  
  public static byte Pack(int limit, int current, boolean player)
  {
    return Pack((byte)limit, (byte)current, player);
  }
  
  /***Restituisce il valore massimo per la casella passata come parametro.***/
  public static byte GetMax(byte cel)
  {
    return (byte)(cel & MASK_MAX);
  }
  
  /***Imposta il valore massimo per la casella passata come parametro e ne ritorna il nuovo valore.***/
  public static byte SetMax(byte cel, byte val)
  {
    return (byte)(val | (cel & (MASK_CUR | MASK_OWN)));
  }
  
  public static byte SetMax(byte cel, int val)
  {
    return SetMax(cel, (byte) val);
  } 
  
  /***Restituisce il valore corrente per la casella passata come parametro.***/
  public static byte GetCurrent(byte cel)
  {
    return (byte)((cel & MASK_CUR) >> 2);
  }
  
  /***Imposta il valore corrente per la casella passata come parametro e ne ritorna il nuovo valore.***/
  public static byte SetCurrent(byte cel, byte val)
  {
    if(val > 4) val = 4;
    return (byte)((val << 2) | (cel & (MASK_MAX | MASK_OWN)));
  }
  
  public static byte SetCurrent(byte cel, int val)
  {
    return SetCurrent(cel, (byte) val);
  }
  
  /***Restituisce il giocatore a cui appartiene la casella passata come parametro.***/
  public static boolean GetOwner(byte cel)
  {
    if(((cel & MASK_OWN) >> 5) == 1) return true;
    return false;
  }
  
  /***Imposta il giocatore proprietario della casella passata come parametro e ne ritorna il nuovo valore.***/
  public static byte SetOwner(byte cel, boolean val)
  {
    byte res;
    if(val) res = 1;
    else res = 0;
    
    return (byte)((res << 5) | (cel & (MASK_MAX | MASK_CUR)));
  }
  
   /***Verifica che la posizione passata sia valida per il giocatore player dato lo stato attuale della partita.***/
  public boolean Verify(int x, int y, boolean player)
  {
    if(x < 0 || x > 5) return false;
    if(y < 0 || y > 4) return false;

    if(GetCelCurrent(x, y) > 0 && GetCelOwner(x, y) != player) return false;
    return true;
  }
  
  /***Verifica che la posizione passata sia valida per il giocatore corrente dato lo stato attuale della partita.***/
  public boolean Verify(int x, int y)
  {
    return Verify(x, y, next_move);
  } 
  
  /***Controlla che non ci sia un vincitore per la partita corrente. Ritorna 0 se non vi sono ancora vincitori, 1 se il giocatore
   � il vincitore e 2 se l'avversario e' il vincitore.***/
  public int CheckForVictory()
  {
    if(turn < 2) return 0; //Troppo presto
    
    if(GetDiscs(true) == 0) return 2; //Ha vinto l'avversario
    if(GetDiscs(false) == 0) return 1; //Ha vinto il giocatore
    
    return 0; //Nessuno ha vinto
  }
  
  /***Aggiunge una pedina del giocatore player alle coordinate (x,y). Ritorna false se la mossa
   specificata non e' valida.***/
  public boolean Add(int x, int y, boolean player) 
  {
    if(!Verify(x, y, player)) return false; //La mossa non e' valida
  
    turn++;
  
    SetCelOwner(x, y, player);
    SetCelCurrent(x, y, GetCelCurrent(x, y) + 1);
    if(GetCelCurrent(x, y) > GetCelMax(x, y)) while(Explode(player)) if(GetDiscs(!player) == 0) break;  
 
    next_move = !next_move;
 
    return true;
  }
  
  /***Aggiunge una pedina del giocatore a cui spetta la mossa alle coordinate (x,y). Ritorna false se la mossa
   specificata non e' valida.***/
  public boolean Add(int x, int y)
  {
    return Add(x, y, next_move);
  }
 
  /***Aggiunge una pedina del giocatore player alle coordinate (x,y) senza controllare la validita' della mossa.***/
  public void AddFast(int x, int y)
  {
    AddFast(x, y, next_move);
  }
  
  /***Aggiunge una pedina del giocatore player alle coordinate (x,y) senza controllare la validita' della mossa.***/
  public void AddFast(int x, int y, boolean player)
  {
    turn++;
    
    SetCelOwner(x, y, player);
    SetCelCurrent(x, y, GetCelCurrent(x, y) + 1);
    if(GetCelCurrent(x, y) > GetCelMax(x, y)) while(Explode(player)) if(GetDiscs(!player) == 0) break; 
  
    next_move = !next_move;
  }
  
  /***Ritorna la casella alla posizione indicata dai parametri.***/
  public byte GetCel(int x, int y)
  {
    return board[x][y];
  }
  
  /***Imposta il proprietario della casella alle coordinate x e y. Non effettua controlli sulla validita' dei parametri.***/
  public void SetCelOwner(int x, int y, boolean owner)
  {
    board[x][y] = SetOwner(board[x][y], owner);
  }
  
  /***Restituisce il proprietario della casella alle coordinate sono x e y. Non effettua controlli sulla validita' dei parametri.***/
  public boolean GetCelOwner(int x, int y)
  {
    return GetOwner(board[x][y]);
  }
  
  /***Imposta il valore corrente della cella alle coordinate sono x e y. Non effettua controlli sulla validita' dei parametri.***/
  public void SetCelCurrent(int x, int y, int val)
  {
    if(val > 4) val = 4;
    board[x][y] = SetCurrent(board[x][y], val);
  }
  
  /***Restituisce il valore corrente della casella alle coordinate x e y. Non effettua controlli sulla validita' dei parametri.***/
  public int GetCelCurrent(int x, int y)
  {
    return (int)GetCurrent(board[x][y]);
  }
  
  /***Imposta il valore massimo della casella alle coordinate x e y. Non effettua controlli sulla validita' dei parametri.***/
  public void SetCelMax(int x, int y, int val)
  {
    board[x][y] = SetMax(board[x][y], val);
  }
  
  /***Restituisce il valore massimo della casella alle coordinate x e y. Non effettua controlli sulla validita' dei parametri.***/
  public int GetCelMax(int x, int y)
  {
    return (int)GetMax(board[x][y]);
  }

  public boolean Explode(boolean player)
  {   
    boolean res = false;
    //Incominciamo il ciclo delle esplosioni
    for(int x = 0; x <= 5; x++)
    {  
      for(int y = 0; y <= 4; y++)
      {
        if(GetCelCurrent(x, y) <= GetCelMax(x, y)) continue;
        res = true;
       
        SetCelCurrent(x, y, 0);
        //Distribuiamo le pedine
        if(x - 1 >= 0)
        {
          SetCelCurrent(x-1, y, GetCurrent(board[x-1][y]) + 1);
          SetCelOwner(x-1, y, player);
        }
        if(x + 1 <= 5)
        {
          SetCelCurrent(x+1, y, GetCelCurrent(x+1, y) + 1);
          SetCelOwner(x+1, y, player);
        }
        if(y - 1 >= 0)
        {
          SetCelCurrent(x, y-1, GetCelCurrent(x, y-1) + 1);
          SetCelOwner(x, y-1, player);
        }
        if(y + 1 <= 4)
        {
          SetCelCurrent(x, y+1, GetCelCurrent(x, y+1) + 1);
          SetCelOwner(x, y+1, player);
        }
      }    
    }
    return res;
  }
  
  /***Rimuove n dischi dalla casella di coordinate (x,y). Se n == 0 allora tutti i dischi verranno tolti.***/
  public boolean Remove(int x, int y, int n) //n == 0 significa 'tutte'
  { 
    if(n == 0) n = 3;
    SetCelCurrent(x, y, GetCelCurrent(x, y) - n);
    if(GetCelCurrent(x, y) < 0) SetCelCurrent(x, y, 0);
  
    if(GetCelCurrent(x, y) > 0) return true;
    return false;
  }
  
  /***Visualizza un messaggio che indica la vittoria del giocatore passato come parametro.***/
  public void PrintVictory(boolean player, boolean mode)
  {
    System.out.println("-----------------------------------\n");
     if(mode == true)
    {
      if(player == true) System.out.println("Partita vinta!");
      else System.out.println("Partita persa!");
    }  
    else
    {
      if(player == true) System.out.println("Ha vinto giocatore 1!");
      else System.out.println("Ha vinto giocatore 2");
    }
    System.out.println("\n-----------------------------------");
  } 
  
  /***Reinizializza lo stato del tabellone di gioco.***/
  public void Reinitialize()
  {
    for(int x = 0; x <= 5; x++)
      for(int y = 0; y <= 4; y++)
    {
      if(x == 0 || x == 5)
      {
        if(y == 0 || y == 4) board[x][y] = Pack(1, 0, false);
        else board[x][y] = Pack(2, 0, false);
      } 
      else if(y == 0 || y == 4) board[x][y] = Pack(2, 0, false);
      else board[x][y] = Pack(3, 0, false);
    }
    turn = 0;
    next_move = false;
  }
  
  /***Restituisce il numero di pedine appartenenti al giocatore specificato.***/
  public int GetDiscs(boolean player)
  {
    int n = 0;
    for(int x = 0; x <= 5; x++)
      for(int y = 0; y <= 4; y++)
      if(GetCelOwner(x, y) == player) n += GetCelCurrent(x, y);
    
    return n;
 }
}
    
    
      