//Gui.java

/***
 Questa classe si occupa della visualizzazione degli elementi grafici dell'interfaccia e della
 gestione dell'input dell'utente. La gestione della connessione tramite socket e' delegata ad
 un'altra classe, ma viene utilizzata da Gui ad un livello piu' alto.

 Autore: Alessandro Caviola
***/

import java.io.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.text.*;
import java.awt.*;
import java.awt.event.*;

class Gui extends JFrame
{
  protected Container container;
  protected Screen screen;
  protected Listener listener = new Listener();
  protected Net net1;
  protected Net net2;
  protected Process process;
  protected Timer timer;
  
  protected int state; /***Specifica lo stato attuale del gioco. Assume uno dei valori delle costanti GAME_STATE_*.***/
  protected int game; /***Specifica il tipo di gioco. Assume uno dei valori delle costanti GAME_TYPE_*.***/
  protected int role; /***Specifica il ruolo della GUI. Assume uno dei valori delle costanti GAME_ROLE_*.***/
  
  protected boolean nextstep;
  protected boolean waiting; /***Se true, indica che il gioco sta aspettando la risposta di un messaggio spedito.***/
  
  protected Board prev = null; /***Copia dello stato della partita prima della precedente mossa.***/
  
  //Costanti
  public static int GAME_STATE_WAIT = 0; /***Il gioco e' in attesa.***/
  public static int GAME_STATE_CONNECTING = 1; /***Il programma sta effettuando una connessione.***/
  public static int GAME_STATE_GAME = 2; /***Il gioco e' in esecuzione.***/
  public static int GAME_STATE_END = 3; /***La partita e' terminata.***/
  public static int GAME_STATE_CONNECTED = 4; /***La partita e' terminata.***/
  
  public static int GAME_TYPE_ONEGUI = 0; /***Due giocatori con la stessa GUI.***/
  public static int GAME_TYPE_GUIAI = 1; /***Un giocatore contro AI.***/
  public static int GAME_TYPE_NETGAME = 2; /***Due giocatori via rete.***/
  
  public static int GAME_ROLE_GAME = 0; /***Funzionamento come server di gioco.***/
  public static int GAME_ROLE_BRIDGE = 1; /***Funzionamento come bridge.***/
  
  //Informazioni
  protected JLabel turn;
  protected JLabel turn_n; //Indica il numero di turni trascorso
  protected JLabel amount_p;
  protected JLabel amount_pn; //Indica il numero di pedine del giocatore
  protected JLabel amount_o;
  protected JLabel amount_on; //Indica il numero di pedine dell'avversario
  protected JLabel player1; 
  protected JLabel playerc; //Indica il colore del giocatore 1
  
  
  //Pulsanti
  protected JButton mode; //Indica la modalita' di funzionamento del programma.
  protected JButton options; //Pannello delle opzioni
  protected JButton connect; //Effettua la connessione
  protected JButton startgame; //Inizia la partita 
  protected JButton quitgame; //Termina partita
  protected JButton next; //Prossimo turno di gioco
  protected JButton withpc; //Avvia un server e ci si connette automaticamente
  protected JButton withpl; //Permette una partita a due su un pc
  protected JButton about; //Visualizza le informazioni sul programma
  
  //Stato del mouse sul piano di gioco
  protected boolean on;
  protected int x, y;
 
  //Variabili delle opzioni

  //Connection options
  protected boolean firstplayer; /***Indica a quale giocatore va la prima mossa.***/
  //Parametri di connessione con il primo server
  protected String addr1;
  protected int port1;
  protected int port12;
  //Parametri di connessione con il secondo server
  protected String addr2;
  protected int port2;
  protected int port22;
  
  protected boolean stepbystep; /***Specifica se la modalita' passo-passo deve essere attivata.***/
  
  //Visual & game options
  protected boolean ex_prot; //Indica se deve essere usata la versione estesa del protocollo
  protected String termch;
  protected int termchi;
  protected boolean ignoreremote; //Ignora le impostazioni remote
  protected int plclr;
  protected int oppclr;
  protected int tbl;
  
  /***Costruisce un oggetto del tipo GUI.***/
  Gui(String title)
  {
    super(title);

    game = GAME_TYPE_ONEGUI;
    role = GAME_ROLE_GAME;
 
    ex_prot = true;
    termchi = 3;
    termch = new String("\\0"); //Fine stringa per dividere i messaggi
    
    firstplayer = true;
    addr1 = new String("127.0.0.1");
    port1 = 1976;
    port12 = 1981;
    addr2 = new String("127.0.0.1");
    port2 = 2002;
    port22 = 2007;
    stepbystep = false;
    
    timer = new Timer(100, new TimerListener());
 
    CreateGUI();
    
    plclr = screen.board.GetPlayerColor(true);
    oppclr = screen.board.GetPlayerColor(false);
    tbl = screen.board.GetBoard();
    
    Initialize();

  }
 
  /***Imposta il programma in uno stato neutro.***/
  public void Initialize()
  {
    state = GAME_STATE_WAIT;
    
    if(timer != null) timer.stop(); //Fermiamo l'eventuale timer
    if(net1 != null) { net1.CloseConnection(); net1 = null; }
    if(net2 != null) { net2.CloseConnection(); net2 = null; }

    if(process != null) { process.destroy(); process = null; }//Chiudiamo il processo
   
    screen.board.Reinitialize();
    screen.board.SetPlayersColors(plclr, oppclr);
    screen.board.SetBoard(tbl);
    System.gc();
    UpdateInfo();
    Enable();
    screen.repaint();
    nextstep = false;
  }
  
  /***Crea l'interfaccia grafica dell'applicazione.***/
  public void CreateGUI()
  {
    container = getContentPane();
    container.setLayout(null);
    
    addWindowListener(listener);
    
    screen = new Screen();
    screen.setBounds(100, 20, 600, 525);
    container.add(screen); 

    mode = new JButton("Game");
    mode.setBounds(0, 20, 100, 30);
    mode.addActionListener(listener);
    container.add(mode);
    
    options = new JButton("Options");
    options.setBounds(0, 50, 100, 30);
    options.addActionListener(listener);
    container.add(options);
 
    connect = new JButton("Connect");
    connect.setBounds(0, 80, 100, 30);
    connect.addActionListener(listener);
    container.add(connect);
    
    startgame = new JButton("Start Game");
    startgame.setBounds(0, 110, 100, 30);
    startgame.addActionListener(listener);
    container.add(startgame);
    
    quitgame = new JButton("Quit Game");
    quitgame.setBounds(0, 140, 100, 30);
    quitgame.addActionListener(listener);
    container.add(quitgame);
    
    next = new JButton("Next Step");
    next.setBounds(0, 200, 100, 30);
    next.addActionListener(listener);
    container.add(next);
    
    withpc = new JButton("Run AI Server");
    withpc.setBounds(0, 260, 100, 30);
    withpc.addActionListener(listener);
    container.add(withpc);
    
    withpl = new JButton("Quick Play");
    withpl.setBounds(0, 290, 100, 30);
    withpl.addActionListener(listener);
    container.add(withpl);
    
    about = new JButton("About");
    about.setBounds(0, 350, 100, 30);
    about.addActionListener(listener);
    container.add(about);
    
    //Informazioni stato partita
    turn = new JLabel("Turno:");
    turn.setBounds(100, 545, 40, 20);
    container.add(turn);
    
    turn_n = new JLabel();
    turn_n.setBounds(140, 545, 30, 20);
    container.add(turn_n);
    
    amount_p = new JLabel("Pedine del giocatore:");
    amount_p.setBounds(190, 545, 130, 20);
    container.add(amount_p);
    
    amount_pn = new JLabel();
    amount_pn.setBounds(320, 545, 30, 20);
    container.add(amount_pn);
    
    amount_o = new JLabel("Pedine dell'avversario:");
    amount_o.setBounds(360, 545, 130, 20);
    container.add(amount_o);
    
    amount_on = new JLabel();
    amount_on.setBounds(490, 545, 30, 20);
    container.add(amount_on);
    
    player1 = new JLabel("Colore giocatore 1:");
    player1.setBounds(540, 545, 120, 20);
    container.add(player1);
    
    player1 = new JLabel();
    player1.setBounds(660, 545, 50, 20);
    container.add(player1);
    
    Enable();
    setResizable(false);
    setSize(800, 600);
    setVisible(true);
  }
  
  /***Imposta il corretto stato degli elementi grafici per l'attuale stato del programma.***/
  public void Enable()
  {
    mode.setEnabled(state == GAME_STATE_WAIT);
    startgame.setEnabled(state == GAME_STATE_CONNECTED);
    quitgame.setEnabled(state != GAME_STATE_WAIT);
    next.setEnabled(role == GAME_ROLE_BRIDGE && state == GAME_STATE_GAME && stepbystep);
    withpc.setVisible(false);// withpc.setEnabled(role == GAME_ROLE_GAME && state == GAME_STATE_WAIT);
    withpl.setEnabled(role == GAME_ROLE_GAME && state == GAME_STATE_WAIT);
    connect.setEnabled(state == GAME_STATE_WAIT);
  }
  
  /***Interrpmpe l'esecuzione del programma per il periodo di tempo specificato in millisecondi.***/
  public void Pause(long millis)
  {
    try
    {
      Thread.sleep(millis);
    }catch(InterruptedException e)
    {
    }
  }

  /***Termina in maniera 'pulita' l'esecuzione del programma.***/
  public void Exit(int code)
  {
    String msg; 
    //Spedisce i messaggi di chiusura
    if(net1 != null && net1.IsActive()) 
    { 
      msg = net1.CreateMsg(Net.MSG_END, 0, 0);
      net1.SendMessage(msg); 
      if(net1.GetError() != Net.NET_ERROR_OK) Error(net1.GetErrorStr()); 
    }
   
    if(role == GAME_ROLE_BRIDGE && net2 != null && net2.IsActive()) 
    {
      msg = net1.CreateMsg(Net.MSG_END, 0, 0);
      net2.SendMessage(msg);     
      if(net2.GetError() != Net.NET_ERROR_OK) Error(net2.GetErrorStr());
    }
      
    Pause(250);
    if(net1 != null) net1.CloseConnection(); //Terminiamo la connessione di rete
    if(net2 != null) net2.CloseConnection();
       
    if(timer != null) timer.stop(); //Fermiamo l'eventuale timer
    
    System.exit(code);
  }
  
  /***Controlla una situazione di vittoria o di perdita ed eventualmente visualizza un messaggio.***/
  protected void Victory()
  {
    int tmp = screen.board.CheckForVictory();
   
    if(tmp == 0) return; //Nessun vincitore, ancora
    
    if(game == GAME_TYPE_ONEGUI || role == GAME_ROLE_BRIDGE)
    {
      if(tmp == 1) screen.board.PrintVictory(true, false);
      else screen.board.PrintVictory(false, false);    
    }
    else
    {
      if(tmp == 1) screen.board.PrintVictory(true, true);
      else screen.board.PrintVictory(false, true);
    }
    
    state = GAME_STATE_END;
    screen.repaint();
  }
  
  /***Esegue i passi necessari per processare il messaggio msg. Il valore ritornato e' il codice del messaggio o -1 in caso di errore.
  Questo metodo puo' impostare lo stato a GAME_STATE_END in seguito ad alcuni eventi (come la vittoria o la perdita. Il chiamante dovrebbe
  controllare quindi lo stato ogni volta che invoca ProcessMsg(). I parametri specificano il messaggio ed il mittente. Questo metodo si 
  occupa di spedire eventuali messaggi di risposta (come MSG_OK).***/
  protected int ProcessMsg(String msg, boolean sender)
  {
    if(msg == null) return -1;
    
    int code = Net.GetCode(msg, termch);
    if(code == -1) return -1; //Codice non valido
    
    //Processiamo il messaggio
    else if(code == Net.MSG_OK)
    {
      Victory();
    }
    else if(code == Net.MSG_FIRST_MOVE) //Incominciamo a giocare
    {
      if(state == GAME_STATE_CONNECTED)
      {
        state = GAME_STATE_GAME;
        game = GAME_TYPE_NETGAME;
        Enable();
      }
      
      screen.board.SetNextMove(true);
      waiting = false;
    }
    else if(code == Net.MSG_MOVE) //Mossa
    {
      if(state == GAME_STATE_CONNECTED)
      {
        state = GAME_STATE_GAME;
        game = GAME_TYPE_NETGAME;
        Enable();
      }
      
      if(!screen.board.Add(Character.getNumericValue(msg.charAt(0)), Character.getNumericValue(msg.charAt(1))))
      {
        //Il messaggio ricevuto non specifica una mossa valida      
        if(ex_prot) //Se usiamo il protocollo esteso, inviamo il messaggio di errore
        {
          String err = Net.CreateMsg(Net.MSG_ERR, 0, 0, termch);
          if(sender == true) { if(!net1.SendMessage(err)) Error(net1.GetErrorStr()); }
          else if(sender == false) { if(!net2.SendMessage(err)) Error(net2.GetErrorStr()); }
          return -1;
        }
        else //Altrimenti interrompiamo il gioco
        {
          Error("Il messaggio ricevuto specifica una mossa non valida: la partita verra' terminata!");
          state = GAME_STATE_END;
          return -1;
        }
      }      
      //Inviamo la conferma al destinatario
      if(ex_prot)
      {
        //Invio la conferma all'avversario
        String err = Net.CreateMsg(Net.MSG_OK, 0, 0, termch);
        if(sender == true)
        {
          if(!net1.SendMessage(err)){ Error(net1.GetErrorStr()); return -1; }
        }
        else if(sender == false)
        {
          if(!net2.SendMessage(err)){ Error(net2.GetErrorStr()); return -1; }
        }
      }
      //Aggiorniamo il gioco
      screen.repaint();
      UpdateInfo();
      
      Victory(); //Controlliamo se c'e' un vincitore
    }
    else if(code == Net.MSG_ERR) //La nostra ultima mossa era non valida
    {
      //NOTA: Questa situazione non dovrebbe verificarsi. Prevista solo per robustezza!
      screen.board.CopyFrom(prev); //Sostituiamo Board con il precedente stato.
      Error("L'applicazione remota ha indicato che l'ultima mossa effettuata non era valida!");
      screen.repaint();
      waiting = false;
      UpdateInfo();
    }
    else if(code == Net.MSG_COLOR) //Indica il colore dell'avversario ed, eventualmente, il colore da assumere
    {
      if(state == GAME_STATE_CONNECTED)
      {
        state = GAME_STATE_GAME;
        game = GAME_TYPE_NETGAME;
        Enable();
      }     
      if(!ex_prot) return code; //Non utilizzato nella versione base
      if(ignoreremote) return code; //Non utilizziamo le impostazioni remote
      screen.board.SetPlayersColors(Net.GetParam(msg, false, termch), Net.GetParam(msg, true, termch));  
      UpdateInfo();
    }
    else if(code == Net.MSG_INFO) //Informazioni sul processo remoto
    {
      if(!ex_prot) return code; //Non supportato nella versione base
      return code;
    }
    else if(code == Net.MSG_END) //Il processo remoto e' terminato
    {
      state = GAME_STATE_END;
    }    
    return code;
  } 
  
  /***Restituisce un riferimento alla finestra corrente.***/
  public JFrame GetFrame()
  {
    return this; 
  }
  
  /***Aggiorna gli elementi dell'interfaccia che visualizzano le informazioni testuali sullo stato della partita.***/
  public void UpdateInfo()
  {
    //Aggiorniamo le informazioni visualizzate
    turn_n.setText(Integer.toString(screen.board.GetTurn()));
    amount_pn.setText(Integer.toString(screen.board.GetDiscs(true)));
    amount_on.setText(Integer.toString(screen.board.GetDiscs(false))); 
    String s = null;
    int clr = screen.board.GetPlayerColor(true);
    if(clr == XBoard.COLOR_RED) s = new String("Rosso");
    else if(clr == XBoard.COLOR_BLU) s = new String("Blu");
    else if(clr == XBoard.COLOR_GREEN) s = new String("Verde");
    else if(clr == XBoard.COLOR_YELLOW) s = new String("Giallo");
    else if(clr == XBoard.COLOR_WHITE) s = new String("Bianco");
    else if(clr == XBoard.COLOR_BLACK) s = new String("Nero");
    
    player1.setText(s);
  }
  
  /***Visualizza una semplice finestra di errore per avvertire l'utente, e scrive lo stesso errore sul flusso di errore.***/
  public void Error(String error)
  {
    JOptionPane.showMessageDialog(this, error, "Error", JOptionPane.ERROR_MESSAGE);
    System.err.println("Errore: " + error);
  }
 
   /***Visualizza una semplice finestra di avvertimento per informare l'utente.***/
  public void Message(String message)
  {
    JOptionPane.showMessageDialog(this, message, "Message", JOptionPane.ERROR_MESSAGE);
  }
  
  /***Implementa la finestra sulla quale verra' disegnata la scacchiera.***/
  class Screen extends Canvas
  {
    protected BevelBorder border = new BevelBorder(BevelBorder.LOWERED);
    public XBoard board;
    
    Screen()
    {
      board = new XBoard(this);
      addMouseListener(listener);
      addMouseMotionListener(listener);
      invalidate();
    }
    
    public void paint(Graphics g)
    { 
      board.Paint(g);
      if(on  && state == GAME_STATE_GAME) 
        if(board.GetNextMove() || game == GAME_TYPE_ONEGUI) board.PaintDisc(g, x, y);
      border.paintBorder(this, g, 0, 0, 600, 525);
    }
  }
  
  /***Implementa i metodi per la reazione all'input dell'utente.***/
  class Listener extends WindowAdapter implements ActionListener, MouseListener, MouseMotionListener
  {
    /***Implementa il codice per la reazione all'input dell'utente e la sua interazione con gli elementi dell'interfaccia
    grafica.***/
    public void actionPerformed(ActionEvent e)
    {
      String command = e.getActionCommand();
      
      //Eseguiamo il comando dell'utente
      if(command == "Exit") Exit(0); 
      
      ///////////////Tipo di modalita'///////////////
      else if(command == "Bridge") //Si passa da Bridge a Game
      {
        mode.setText("Game");
        role = GAME_ROLE_GAME;
        Enable();
      }
      else if(command == "Game") //Si passa da Game a Bridge
      {
        mode.setText("Bridge");
        role = GAME_ROLE_BRIDGE;
        Enable();
      }

      ///////////////Opzioni///////////////
      else if(command == "Options") //Mostra il pannello delle opzioni
      {
        Options opt = new Options(GetFrame());
        
        if(firstplayer == true) opt.player1.setSelected(true);
        else opt.player2.setSelected(true);
        
        opt.ipctrl1.setText(addr1);
        opt.portctrl1.setText(Integer.toString(port1));
        opt.portctrl12.setText(Integer.toString(port12));
        
        opt.ipctrl2.setText(addr2);
        opt.portctrl2.setText(Integer.toString(port2));
        opt.portctrl22.setText(Integer.toString(port22));
        
        opt.stepby.setSelected(stepbystep);
        
        opt.playerc.select(screen.board.GetPlayerColor(true));
        opt.opponentc.select(screen.board.GetPlayerColor(false));
        
        opt.tablelist.select(screen.board.GetBoard());        
        opt.ext_prot.setSelected(ex_prot);
        opt.terminator.select(termchi);
        
        opt.ignorerem.setSelected(ignoreremote);
        
        opt.setVisible(true);
        if(opt.exit == false) return;
        
        if(opt.player1.isSelected() == true) firstplayer = true;
        else firstplayer = false;
        
        addr1 = opt.ipctrl1.getText();
        port1 = Integer.parseInt(opt.portctrl1.getText());
        port12 = Integer.parseInt(opt.portctrl12.getText());
        
        addr2 = opt.ipctrl2.getText();
        port2 = Integer.parseInt(opt.portctrl2.getText());
        port22 = Integer.parseInt(opt.portctrl22.getText());
        
        firstplayer = opt.player1.isSelected(); 
        
        stepbystep = opt.stepby.isSelected();
        
        plclr = opt.playerc.getSelectedIndex();
        oppclr = opt.opponentc.getSelectedIndex();
        screen.board.SetPlayersColors(plclr, oppclr);
        
        tbl = opt.tablelist.getSelectedIndex();
        screen.board.SetBoard(tbl);
            
        if(state != GAME_STATE_WAIT)
        {
          if(opt.ext_prot.isSelected() != ex_prot || opt.ignorerem.isSelected() != ignoreremote || opt.terminator.getSelectedIndex() != termchi)
            Message("I parametri relativi al protocollo non possono essere modificati durante la partita!");
          
          return;
        }
        ex_prot = opt.ext_prot.isSelected();
        termchi = opt.terminator.getSelectedIndex();
        
        if(!ex_prot) termch = null;        
        else termch = new String(opt.terminator.getItem(termchi));
        
        ignoreremote = opt.ignorerem.isSelected();
        
        Enable();    
        UpdateInfo();
      }
      
      ///////////////Chiusura della partita///////////////
      else if(command == "Quit Game") //Termina il gioco corrente
      {
        String msg = Net.CreateMsg(Net.MSG_END, 0, 0, termch);
        //Spedisce i messaggi di chiusura
        if(net1 != null && net1.IsActive()) 
        { 
          net1.SendMessage(msg); 
          if(net1.GetError() != Net.NET_ERROR_OK) Error(net1.GetErrorStr()); 
        }
        
        if(role == GAME_ROLE_BRIDGE && net2 != null && net2.IsActive()) 
        {
          net2.SendMessage(msg);     
          if(net2.GetError() != Net.NET_ERROR_OK) Error(net2.GetErrorStr());
        }
        
        Pause(250);
        if(timer != null) timer.stop(); //Fermiamo l'eventuale timer
        Initialize();
      }   
    
      ///////////////Esegue il server AI///////////////
      else if(command == "Run AI Server") //Esegue il server AI 
      {
/*        if(state != GAME_STATE_WAIT) //Una partita e' gia' in esecuzione
        {
          Message("Chiudere prima la partita corrente!");
          return;
        }
        
        state = GAME_STATE_CONNECTING;
        game = GAME_TYPE_GUIAI;
        role = GAME_ROLE_GAME;
        timer.start();
        try
        {
          process = Runtime.getRuntime().exec("java AIServer -porti " + port1 + " -porto " + port2 + " -onetime"); //Lancia AIServer
        }catch(IOException ex)
        {
          Error("Impossibile eseguire AIServer!");
          timer.stop();
          Initialize();
          return;
        }
        
        //Connettiamoci al nuovo processo
        net = new Net(addr1, port1, port12);
        net.SetTerm(termch);
        net.Connect();
        
        if(firstmove == false)
        {
          screen.board.SetTurn(false);
          
          String msg = net.CreateMsg(Net.MSG_FIRST_MOVE, 0, 0);
          if(net.SendMessage(msg, true) == false)
          {
            //In questo punto bisogna interrompere il gioco
            Error(net.GetErrorStr());
            net.CloseConnection();
            net = null;
            Initialize();
            return;
          }          
        }
        else screen.board.SetTurn(true);
        Enable();
    
        timer.start();
        return;*/
      } 
      
      ///////////////Due giocatori con la stessa GUI///////////////
      else if(command == "Quick Play") //Inizia una partita a due giocatori
      {
        state = GAME_STATE_GAME;
        game = GAME_TYPE_ONEGUI;
        role = GAME_ROLE_GAME;
        
        screen.board.SetNextMove(true);
        Enable();
        return;
      }
      
      ///////////////Connette l'applicazione///////////////
      else if(command == "Connect") //Connette il programma
      {
        state = GAME_STATE_CONNECTING;
        
        //Connettiamo il primo canale
        net1 = new Net(addr1, port1, port12);
        net1.SetTerm(termch);
        net1.Connect();
       
        if(role == GAME_ROLE_BRIDGE) { net2 = new Net(addr2, port2, port22); net2.SetTerm(termch); net2.Connect(); }
        
        timer.start();
        Enable();
      }
      
      ///////////////Da il via ad una partita in rete///////////////
      else if(command == "Start Game") //Da inizio ad una partita in rete. Si suppone che la connessione sia gia' avvenuta con successo
      {
        state = GAME_STATE_GAME;
        game = GAME_TYPE_NETGAME;
        
        String msg;
        if(net1 != null) if(net1.Ready()) return; //Ci sono gia' messaggi in arrivo: un altro bridge?
        if(net2 != null) if(net2.Ready()) return; //Ci sono gia' messaggi in arrivo: un altro bridge?
        
        screen.board.SetNextMove(firstplayer);
        
        if(role == GAME_ROLE_GAME) //Modalita' gioco
        {
          if(ex_prot)
          {
            msg = Net.CreateMsg(Net.MSG_COLOR, screen.board.GetPlayerColor(true), screen.board.GetPlayerColor(false), termch);
            if(net1.SendMessage(msg) == false)
            {
              Error(net1.GetErrorStr());
              Initialize();
              return;
            }  
          }
          //Spediamo il messaggio a chi dovra' iniziare
          msg = Net.CreateMsg(Net.MSG_FIRST_MOVE, 0, 0, termch);
          if(!firstplayer)
          {
            if(!net1.SendMessage(msg))
            {
              Error(net1.GetErrorStr());
              Initialize();
              return;
            }   
          }  
          Enable();
          return;
        }
          
        //Spediamo il messaggio indicante i colori usati
        if(ex_prot)
        {             
          msg = Net.CreateMsg(Net.MSG_COLOR, screen.board.GetPlayerColor(false), screen.board.GetPlayerColor(true), termch);
          if(net1.SendMessage(msg) == false)
          {
            Error(net1.GetErrorStr());
            Initialize();
            return;
          }  
          msg = Net.CreateMsg(Net.MSG_COLOR, screen.board.GetPlayerColor(true), screen.board.GetPlayerColor(false), termch);
          if(net2.SendMessage(msg) == false)
          {
            Error(net2.GetErrorStr());
            Initialize();
            return;
          }  
        }
        
        //Spediamo il messaggio a chi dovra' iniziare
        msg = Net.CreateMsg(Net.MSG_FIRST_MOVE, 0, 0, termch);
        if(firstplayer)
        {
          if(!net1.SendMessage(msg))
          {
            Error(net1.GetErrorStr());
            Initialize();
            return;
          }   
        }
        else
        {
          if(!net2.SendMessage(msg))
          {
            Error(net2.GetErrorStr());
            Initialize();
            return;
          }   
        }
        
        Enable();
        return;
      }
      
      ///////////////Visualizza una finestra informativa///////////////
      else if(command == "About") //Visualizza una finestra di informazioni sul programma
      {
        String s = new String("Massa Critica v.1.3\n\nAlessandro Caviola 2004");
        JOptionPane.showMessageDialog(null, s, "About", JOptionPane.ERROR_MESSAGE);  
      }
      
      ///////////////Permette l'attuazione della mossa successiva///////////////
      else if(command == "Next Step") //Permette l'esecuzione della prossima mossa
      {
        nextstep = true;
      }
    }
    
    public void windowClosing(WindowEvent e)
    {
      Exit(0);
    }
    
    public void mouseClicked(MouseEvent e)
    {
      if(state != GAME_STATE_GAME || role == GAME_ROLE_BRIDGE) return; //Non si devono intraprendere azioni
      if(!screen.board.GetNextMove() && game != GAME_TYPE_ONEGUI) return; 
      
      //Recupero le coordinate del 'click' e le trasformo nelle coordinate della casella specificata
      Point p = screen.board.GetHitCoords(e.getX(), e.getY());
      if(p == null) return;
      
      if(!screen.board.Add(p.x, p.y)) //Aggiungiamo, se possibile, una pedina alla casella specificata
      {
        Error("Mossa non valida!"); //La mossa specificata non e' valida
        return;
      }   
      screen.repaint(); //Ridisegnamo tutto per via delle possibili esplosioni
      if(game != GAME_TYPE_ONEGUI)
      {
        String tmp = Net.CreateMsg(Net.MSG_MOVE, p.x, p.y, termch);
        if(!net1.SendMessage(tmp))
        {
          Error(net1.GetErrorStr());
          return;
        }
        waiting = true; //Questo messaggio necessita di una risposta. ATTENZIONE: e' necessario che tutti i programmi coinvolti
                        //abbiano attivata l'opzione Estensione protocollo o che nessuno l'abbia attivata.
      }  

      UpdateInfo();
      
      Victory();
    }
    
    public void mouseEntered(MouseEvent e)
    {
      on = true;
    }
    
    public void mouseExited(MouseEvent e)
    {
      on = false;

      if(state != GAME_STATE_GAME || role == GAME_ROLE_BRIDGE) return; //Non si devono intraprendere azioni
      if(!screen.board.GetNextMove() && game != GAME_TYPE_ONEGUI) return;  
      
      int _x = x;
      int _y = y;
      x = e.getX();
      y = e.getY();
      screen.repaint(_x-40, _y-40, 80, 80);
    }
    
    public void mousePressed(MouseEvent e)
    {
    }
    
    public void mouseReleased(MouseEvent e)
    {
    }
    
    public void mouseMoved(MouseEvent e)
    {
      int _x = x;
      int _y = y;
      x = e.getX();
      y = e.getY();
      
      if(state != GAME_STATE_GAME || role == GAME_ROLE_BRIDGE) return; //Non si devono intraprendere azioni
      if(!screen.board.GetNextMove() && game != GAME_TYPE_ONEGUI) return;  
      
      screen.repaint(_x-40, _y-40, 80, 80);
    }
    
    public void mouseDragged(MouseEvent e)
    {
    }   
  }
  
  class TimerListener implements ActionListener
  {
    /***Risponde alla periodica chiamata del timer ricevendo e spedendo i messaggi. Aggiorna lo stato del gioco in funzione
    dei messaggi ricevuti e verifica la vittoria dei giocatori.***/
    public void actionPerformed(ActionEvent e)
    {
      if(net1 == null) { timer.stop(); return; }
      if(role == GAME_ROLE_BRIDGE) if(net2 == null) { timer.stop(); return; } 
      
      if(net1.GetError() != Net.NET_ERROR_OK) //Errore durante la connessione
      {
        Error(net1.GetErrorStr());
        if(net1.GetError() == Net.NET_ERROR_NOTACTIVE) return;
        
        if(role == GAME_ROLE_BRIDGE) //Spedisco un messaggio di termine, se necessario
        {
          String msg = Net.CreateMsg(Net.MSG_END, 0, 0, termch); 
          net1.SendMessage(msg);
          net2.SendMessage(msg);
        }
        
        Error("La partita verra' ora terminata!"); //Unico modo per recuperare da questo errore
        Pause(250);
        Initialize();
        return;
      }
      
      //Siamo connessi ma il gioco non e' ancora iniziato
      if(state == GAME_STATE_CONNECTING && net1.IsActive())
      {
        if(role == GAME_ROLE_BRIDGE && !net2.IsActive()) return; //Il bridge necessita di entrambe le connessioni
        state = GAME_STATE_CONNECTED;
        Enable();
        return;
      }
      
      if(role == GAME_ROLE_BRIDGE && net2.GetError() != Net.NET_ERROR_OK) //Errore durante la connessione
      {
        Error(net2.GetErrorStr());
        if(net2.GetError() == Net.NET_ERROR_NOTACTIVE) return;
        
        String msg = Net.CreateMsg(Net.MSG_END, 0, 0, termch); 
        net1.SendMessage(msg);
        net2.SendMessage(msg);
        
        Error("La partita verra' ora terminata!"); //Unico modo per recuperare da questo errore
        Pause(250);
        Initialize();
        return;
      }

      ///////
      //Modalita' Game
      ///////
      
      else if(role == GAME_ROLE_GAME) 
      {  
        if(!net1.IsActive()) return; //La connessione non e' ancora pronta
        
        String msg = net1.GetMessage(); //Leggiamo l'eventuale messaggio
        if(msg == null) 
        {
          if(net1.GetError() != net1.NET_ERROR_OK)
            Error(net1.GetErrorStr()); //Errore
          
          return; //Nessun messaggio
        } 
        
        //Processiamo il messaggio
        int code = ProcessMsg(msg, true);
        if(state == GAME_STATE_END)
        {
          if(screen.board.CheckForVictory() != 0) return; //Se qualcuno ha vinto, la chiusura non e' automatica
          if(code == Net.MSG_END)
            Message("Un utente remoto ha abbandonato la partita: la partita verra' ora terminata!");
       
          Initialize();
        }
      }
      
      ////////
      //Modalita' Bridge
      ////////
      
      else if(role == GAME_ROLE_BRIDGE) 
      {
            if(!net1.IsActive() || !net2.IsActive()) return; //La connessione non e' ancora pronta

        if((stepbystep == true && nextstep == true) || stepbystep == false) //Leggiamo dal primo canale
        {
          String msg = net1.GetMessage(); //Leggiamo l'eventuale messaggio
          if(msg == null) //Nessun messaggio o errore
          {
            if(net1.GetError() != net1.NET_ERROR_OK) 
              Error(net1.GetErrorStr()); //Errore
          }
          else
          {
            //Processiamo il messaggio
            int code = ProcessMsg(msg, true);
            if(code == Net.MSG_MOVE) nextstep = false;
            if(code != -1 && code != Net.MSG_OK) net2.SendMessage(msg);
            if(state == GAME_STATE_END)
            {
              if(screen.board.CheckForVictory() != 0) return; //Se qualcuno ha vinto, la chiusura non e' automatica
              if(code == Net.MSG_END)
                Message("Un utente remoto ha abbandonato la partita: la partita verra' ora terminata!");
              
              msg = Net.CreateMsg(Net.MSG_END, 0, 0, termch);
              net2.SendMessage(msg);
              Pause(250);
              
              Initialize();
              return;
            }
          }
        }
        if((stepbystep == true && nextstep == true) || stepbystep == false) //Leggiamo dal secondo canale
        {
          String msg = net2.GetMessage(); //Leggiamo l'eventuale messaggio
          if(msg == null) //Nessun messaggio o errore
          {
            if(net2.GetError() != net2.NET_ERROR_OK) 
              Error(net2.GetErrorStr()); //Errore
          }
          else
          {
            //Processiamo il messaggio
            int code = ProcessMsg(msg, false);
            
            if(code == Net.MSG_MOVE) nextstep = false;
            if(code != -1 && code != Net.MSG_OK) net1.SendMessage(msg);
            if(state == GAME_STATE_END)
            {
              if(screen.board.CheckForVictory() != 0) return; //Se qualcuno ha vinto, la chiusura non e' automatica
              else if(code == Net.MSG_END)
                Message("Un utente remoto ha abbandonato la partita: la partita verra' ora terminata!");
 
              msg = Net.CreateMsg(Net.MSG_END, 0, 0, termch);
              net1.SendMessage(msg);
              Pause(250);
              
              Initialize();
              return;
            }
          }
        }
      } //Bridge
    } //ActionPerformed()
  } //TimerListener
}