//Net.java

/***
 Questa classe si occupa di impostare e gestire la connessione tra i programmi. Oltre a questo, 
 fornisce i metodi per l'utilizzo del protocollo. La Classe Net utilizza un thread separato per effettuare la connessione, e
 un thread distinto per ogni canale I/O.
 
 autore: Alessandro Caviola
***/
import java.util.*;
import java.net.*;
import java.io.*;

class Net extends Thread
{ 
  protected NBIO in; /***Canale di lettura della prima connessione.***/
  protected NBIO out; /***Canale di lettura della seconda connessione.***/
  
  protected Socket socket1, socket2;
  protected ServerSocket s;
  
  protected String address;
  protected int port1, port2;
  
  protected int error; /***Mantiene una valore relativo al successo o al fallimento di alcune operazioni.***/
  protected String term = null;
  protected int timeout; /***Indica quanti millisecondi concedere al massimo ad un tentativo di connessione.***/
  
  protected boolean running;
    
  //Codici per gli errori
  public static int NET_ERROR_OK = 0; /***Nessun errore.***/
  public static int NET_ERROR_CONNECTION = 1; /***Errore durante la connessione.***/
  public static int NET_ERROR_IO = 2; /***Errore nel trasferimento dei dati.***/
  public static int NET_ERROR_MSG = 3; /***Messaggio non valido.***/
  public static int NET_ERROR_NOTACTIVE = 4; /***La connessione non e' attiva.***/
  public static int NET_ERROR_UNKNOWN = 5; /***Si e' verificato un errore non specificato.***/
  public static int NET_ERROR_ALREADYACTIVE = 6; /***Una connessione e' gia' attiva.***/
  public static int NET_ERROR_LOST = 7; /***E' stata persa la connessione.***/
  
  //Codici per i messaggi
  public static int MSG_OK = 60; /***Versione estesa: il messaggio e' corretto =>'OK'.***/
  public static int MSG_FIRST_MOVE = 61; /***Permesso per la prima mossa => '0'.***/
  public static int MSG_MOVE = 62; //Esecuzione di una mossa => 'xy' con 0 <= x <= 5 e 0 <= y <= 4.***/
  public static int MSG_ERR = 63; //Versione estesa: messaggio errato => 'ERR'.***/
  public static int MSG_COLOR = 64; //Versione estesa: indicazione sul colore => 'COLOR'.***/
  public static int MSG_INFO = 65; //Versione estesa: informazioni => 'INFO'.***/
  public static int MSG_END = 66; //Fine partita => 'END'.***/
  
  
  /***Costruisce un oggetto Net. I parametri passati sono le porte e l' indirizzo dei
  server a cui connettersi.***/
  Net(String addr, int p1, int p2)
  {
    in = new NBIO(true);
    out = new NBIO(false);
    
    address = addr; 
    port1 = p1;
    port2 = p2;
    
    socket1 = null;
    socket2 = null;
    
    error = NET_ERROR_OK;

    timeout = 60000; //Un minuto di tempo perche' un server si renda disponibile
    term = new String("");
  }
  
  /***Indica se tutte le connessioni richieste sono attive.***/
  public boolean IsActive()
  {    
    return (in.active && out.active);
  }
  
  /***Restituisce il codice di errore attuale.***/
  public int GetError()
  {
    return error;
  }
 
  /***Restituisce una stringa descrittiva dell'attuale codice di errore.***/
  public String GetErrorStr()
  {
    if(error == NET_ERROR_OK)
      return new String("OK");
    else if(error == NET_ERROR_CONNECTION)
      return new String("Impossibile stabilire una connessione!");
    else if(error == NET_ERROR_IO)
      return new String("Errore duarante la trasmissione o la ricezione dei dati!");
    else if(error == NET_ERROR_MSG)
      return new String("Messaggio non valido!");
    else if(error == NET_ERROR_NOTACTIVE)
      return new String("Connessione non attiva!");
    else if(error == NET_ERROR_ALREADYACTIVE)
      return new String("Una connessione e' gia' attiva!");
    else if(error == NET_ERROR_LOST)
      return new String("La connessione e' stata persa!");
    else
      return new String("Si e' verificato un errore sconosciuto!");
  }
  
  /***Resetta il codice di errore impostandolo a NET_ERROR_OK.***/
  public void ResetError()
  {
    error = NET_ERROR_OK;
  } 

  /***Realizza le connessioni richieste in base ai parametri passati al costruttore. Il valore di ritorno pari a false indica
  un errore nello svolgimento delle operazioni, mentre true specifica che la il tentativo di connessione e' stato avviato regolarmente.
  Attenzione: un ritorno pari a true non significa che la connessione sia avvenuta con successo!***/
  public boolean Connect()
  {
    if(IsActive()) //Una connessione e' gia' attiva
    {
      error = NET_ERROR_ALREADYACTIVE;
      return false;
    }
    
    setPriority(MIN_PRIORITY);
    start();
   
    return true;
  }
  

  /***Chiude qualsiasi connessione (sia attiva che in fase di connessione) ripristinando lo stato delle flags.***/
  public void CloseConnection()
  {
    if(running)
    {
      try
      {
        if(s != null) s.close();
        s = null;
        if(socket1 != null) socket1.close();
        socket1 = null;
        if(socket2 != null) socket2.close();
        socket2 = null;
      }catch(IOException e)
      {}
    }
      
    in.Close();
    out.Close();

    running = false;
  }
  
  /***Stabilisce le connessioni ed esegue i threads per l'I/O.***/
  public void run()
  {
    error = NET_ERROR_OK;
    running = true;
    
    in.Run();
    out.Run();


    running = false;
  }
  
  /***Legge un messaggio da una connessione attiva. Il valore ritornato e' una stringa con il messaggio in caso di successo o null
   se si e' verificato un errore o non ci sono dati disponibili.***/
  public String GetMessage()
  {
    if(!IsActive())
    {
      error = NET_ERROR_NOTACTIVE;
      return null;
    }
    
    try
    {
      if(in.msglist.size() == 0) return null; //Non ci sono elementi nella lista
      
      String tmp = new String((String)in.msglist.getFirst());
      in.msglist.removeFirst();
      return tmp;
      
    }catch(NoSuchElementException e)
    {
      return null;  //Non ci sono messaggi disponibili
    }
  }
  
  /***Scrive un messaggio su una connessione attiva. Il parametri indica il messaggio da inviare. 
  Il valore ritornato e' true in caso di successo o false altrimenti.***/
  public boolean SendMessage(String msg)
  {
    if(!IsActive())
    {
      error = NET_ERROR_NOTACTIVE;
      return false;
    }
    
    out.msglist.addLast(msg); 
       
    error = NET_ERROR_OK;
    return true;
  }
  
  /***Restituisce true se ci sono messaggi pronti per la lettura, e false altrimenti.***/
  public boolean Ready()
  {
    if(in == null) return false;
    return !in.msglist.isEmpty();
  }
  
  /***Imposta il carattere di terminazione per ogni messaggio che verra' creato. Se term = null o = "", allora non verra' aggiunto nessun 
  carattere. Se term ha una lunghezza maggiore di 1, solo il primo carattere verra' preso.***/
  public void SetTerm(String term)
  {
    this.term = term;
    if(this.term == "") this.term = null;
  }
  
  /***Restituisce una stringa contenente il carattere terminante.***/
  public String GetTerm()
  {
    return term;
  }
  
  /***Costruisce un messaggio valido utilizzando i parametri passati. Il parametro code e' il codice MSG_* del messaggio che si desidera 
  creare, mentre gli altri due parametri assumono valori che dipendono dal messaggio. Il metodo ritorna il messaggio generato in caso
  di successo o null se i parametri o il codice passato non sono validi.***/
  public String CreateMsg(int code, int param1, int param2)
  {
    return CreateMsg(code, param1, param2, term);
  }
  
  /***Costruisce un messaggio valido utilizzando i parametri passati. Il parametro code e' il codice MSG_* del messaggio che si desidera 
  creare, mentre gli altri due parametri assumono valori che dipendono dal messaggio. Il metodo ritorna il messaggio generato in caso
  di successo o null se i parametri o il codice passato non sono validi.***/
  static public String CreateMsg(int code, int param1, int param2, String term)
  {
    StringBuffer msg = new StringBuffer("");
 
    if(code == MSG_OK)
      msg.append("OK");
    else if(code == MSG_FIRST_MOVE)
      msg.append("0");
    else if(code == MSG_MOVE)
    {        
      msg.append(Integer.toString(param1));
      msg.append(Integer.toString(param2));
    }
    else if(code == MSG_ERR)
      msg.append("ERR");
    else if(code == MSG_COLOR)
    {  
      msg.append("COLOR ");
      msg.append(Integer.toString(param1));
      msg.append(" ");
      msg.append(Integer.toString(param2));
    }
    else if(code == MSG_INFO)
    {
      msg.append("INFO ");
      msg.append(Integer.toString(param1));
    }
    else if(code == MSG_END)
      msg.append("END");
    //Altrimenti il codice non e' valido e viene restituito null
    else return null;
    
    if(term != null) msg.append(term);
    
    return msg.toString();
  }
  
  /***Restituisce il codice del messaggio dato il messaggio stesso, passato come parametro. Il valore di ritorno e' -1
  se msg e' null o la stringa e' vuota o contiene solo un delimitatore.***/ 
  public int GetCode(String msg)
  {
    return GetCode(msg, term);
  }
  
  /***Restituisce il codice del messaggio dato il messaggio stesso, passato come parametro. Il valore di ritorno e' -1
  se msg e' null o la stringa e' vuota o contiene solo un delimitatore.***/
  static public int GetCode(String msg, String term)
  {
    if(msg == null || msg == "") return -1;
    
    String tmp;
    String t;
    if(term != null && term != "")
      t = msg.substring(0, msg.length()-term.length());
    else t = msg;
    
    StringTokenizer tokens = new StringTokenizer(t);
    if(tokens.hasMoreTokens()) tmp = tokens.nextToken(); //Ritorna il primo token
    else return -1;
    
    if(tmp.compareTo("OK") == 0) return MSG_OK;
    else if(tmp.compareTo("0") == 0) return MSG_FIRST_MOVE;
    else if(tmp.compareTo("ERR") == 0) return MSG_ERR;
    else if(tmp.compareTo("COLOR") == 0) return MSG_COLOR;
    else if(tmp.compareTo("INFO") == 0) return MSG_INFO;
    else if(tmp.compareTo("END") == 0) return MSG_END;
    else //MSG_MOVE
    {
      int n;
      try
      {
        n = Integer.parseInt(tmp);
      }catch(NumberFormatException e)
      {
        return -1; //Il formato non era valido
      }
      
      if(n < 0 || n > 54) return -1; //Deve essere lungo due caratteri
      
      return MSG_MOVE;
    }
  }
  
  /***Restituisce uno dei parametri, se esistente, del messaggio passato come parametro. Il secondo parametro di questo
  metodo indica se si desidera il primo parametro del messaggio (TRUE) o il secondo (FALSE). Il risultato e' il parametro 
  o -1 se non esiste tale parametro nel messaggio.***/
  public int GetParam(String msg, boolean par)
  {
    return GetParam(msg, par, term);
  }
  
  /***Restituisce uno dei parametri, se esistente, del messaggio passato come parametro. Il secondo parametro di questo
  metodo indica se si desidera il primo parametro del messaggio (TRUE) o il secondo (FALSE). Il risultato � il parametro 
  o -1 se non esiste tale parametro nel messaggio.***/
  static public int GetParam(String msg, boolean par, String term)
  {
    if(msg == null || msg == "") return -1;
    
    String tmp;
    String t;
    if(term != null && term != "")
      t = msg.substring(0, msg.length()-term.length());
    else t = msg;
    StringTokenizer tokens = new StringTokenizer(t);
    
    tokens.nextToken();
    if(tokens.hasMoreTokens()) tmp = tokens.nextToken(); //Ritorna il primo token
    else return -1;
    if(!par)
    {
      if(tokens.hasMoreTokens()) tmp = tokens.nextToken(); //Ritorna il secondo token
      else return -1;
    }
    
    int n;
    try
    {
      n = Integer.parseInt(tmp);
    }catch(NumberFormatException e)
    {
      return -1; //Il formato non era valido
    }
    return n;
  }
  
  /***Verifica che la stringa passata come parametro sia un messaggio completo, di senso compiuto. Il secondo parametro � 
  l'eventuale carattere terminatore che si vuole controllare, o null se non ci si aspetta un carattere terminatore.***/
  static public boolean MsgComplete(String msg, String term)
  {
    //Controlliamo se lo header � intatto
    int h = GetCode(msg, term);
    
    if(h != MSG_OK && h != MSG_ERR && h != MSG_INFO && h != MSG_END && h != MSG_COLOR && h != MSG_FIRST_MOVE && h != MSG_MOVE)
      return false; //Non � valido
   
    //Controlliamo gli eventuali parametri
    if(h == MSG_COLOR)
    {
      if(GetParam(msg, true, term) == -1 || GetParam(msg, false, term) == -1) return false; //Non � valido
    }
 
    //Controlliamo il carattere terminatore, se richiesto
    if(term == null || term == "") return true; //Il messaggio � valido
    if(!msg.endsWith(term)) return false; 
    
    return true;
  }
  
  /***Gestisce l'IO non bloccante.***/
  class NBIO extends Thread
  {
      public InputStream in;   
      public OutputStream out;
      public LinkedList msglist = new LinkedList(); //Buffer per i messaggi
      public StringBuffer buffer = new StringBuffer(); //Buffer per i messaggi incompleti
            
      public boolean stop; /***Quando true indica al thread di fermarsi non appena possibile.***/
      public boolean running; /***Quando true indica che il Thread � in esecuzione.***/
      public boolean mode; /***Se true allora � in lettura, se false in scrittura.***/ 
      
      boolean active;
      
      NBIO(boolean m)
      {
        stop = false;
        running = false;
        buffer = new StringBuffer("");
        active = false;
        mode = m;
      }
      
      public boolean Connect()
      {
        try
        {
          InetSocketAddress a;        
          if(mode)
          {
            //Apriamo canale di input
            s = new ServerSocket(port2,1);  
            s.setReuseAddress(true);
         
            socket1 = s.accept();
        
            socket1.setReuseAddress(true);
            s = null;  
            
            //Imposta gli stream associati ai socket
            in = new DataInputStream(socket1.getInputStream());
          }
          else
          {
            //Apriamo canale di output
            if(address.compareTo("localhost") == 0) address = new String("127.0.0.1");
            socket2 = new Socket();
            socket2.setReuseAddress(true);
            a = new InetSocketAddress(address, port1);
         
            socket2.connect(a, timeout);   
  
            //Imposta gli stream associati ai socket
            out = new DataOutputStream(socket2.getOutputStream()); 
          } 
        }
        catch(UnknownHostException e)
        {
          error = NET_ERROR_CONNECTION;         
          CloseConnection();
          running = false;
          return false;
        }
        catch(IOException e)
        {
          return false;
        }
        error = NET_ERROR_OK;
        active = true;  
        
        return true;
      }
      
      public void Run()
      {
        setPriority(MIN_PRIORITY);
        start();
      }
        
      public void run()
      {  
        for(int i = 0; i < timeout/500 && !active;  i++) 
        { 
          Connect();
          if(error != NET_ERROR_OK)
          {
            CloseConnection();
            return;
          }
          try
          {
            sleep(500); 
          }catch(InterruptedException e){}
        }
  
        if(!active) return;
        
        error = NET_ERROR_OK;
        
        running = true;
        
        byte tmp[] = new byte[50];
        try
        {
          //Iniziamo il ciclo di I/O
          while(true)
          {          
            if(mode) //Lettura
            {          
              int l = in.read(tmp);
        
              if(l == -1) //La connessione e' stata persa
              {
                error = NET_ERROR_LOST;
                CloseConnection();
                return;
              }
              buffer.append(new String(tmp, 0, l)); 
              
              //Evidenzia i messaggi nella stringa e li aggiunge alla lista
              StringBuffer t = new StringBuffer();
              int index = 0;
              while(true) //Fino a 
              {
                if(term != null)
                {
                              
                  index = buffer.toString().indexOf(term);
                  if(index == -1) break; //Non ci sono messaggi completi
                  
                  t.append(buffer.substring(0, index+term.length())); 
                }
                else
                    t.append(buffer); 
                
                if(MsgComplete(t.toString(), term))
                {
                  msglist.addLast(t.toString());
                  
                  if(term != null) buffer.delete(0, index+term.length());
                  else buffer.delete(0, buffer.length());
                  t.delete(0, t.length());
                  
                  if(term == null) break; //Se non c'e' terminatore, allora dopo il primo messaggio ci si ferma
                } 
                else break;
              }  
            }
            else //Scrittura
            {
              if(msglist.size() == 0) 
              {
                try
                {
                  sleep(100); //Aspettiamo un po'
                  continue;
                }catch(InterruptedException e)
                {
                }
              }
              byte temp[] = ((String)msglist.getFirst()).getBytes();
              
              out.write(temp);
              msglist.removeFirst();              
            }
            if(stop) //Terminiamo il thread, se richiesto
            {          
              running = false;
              stop = false;
              return;
            }
          } 
        }catch(IOException e)
        {
          CloseConnection();
          error = NET_ERROR_IO;
        }
        running = false;
      }
       
      public void Close()
      {    
        if(!running) return;
        stop = true;    
        active = false;
      }
  } 
}