//Server.java

/***
 Questa classe esegue il lavoro del server AI, come la gestione della connessione e la stampa dei messaggi sulla console.
 
 Autore: Alessandro Caviola
***/
import java.io.*;
import java.awt.event.*;
import java.awt.*;
import javax.swing.*;

class Server
{
  public Net net; 
  public Board board;
  public Board oldb;
  public Movegen generator; /***Permette la generazione delle mosse***/
  public int state; 
  public Timer timer; 
  
  //Parametri
  public boolean verbose;
  public boolean oldp;
  public String term;
  public int port;
  public int port2;
  public String addr;
  public boolean console;
  public int tree_d;
  public int diff;
  public boolean test;
  public boolean onetime;
   
  //Alcune costanti
  public static int SERVER_STATE_CONNECTING = 0;
  public static int SERVER_STATE_GAME = 1;
  public static int SERVER_STATE_END = 2;
  
  Server()
  {
    net = null; 
    board = new Board(); 
    oldb = null;
    
    verbose = false;
    oldp = false;
    port = 1976;
    port2 = 1981;
    term = new String("\\0");
    addr = new String("127.0.0.1");
    console = false;
    tree_d = 3;
    diff = 0;
    test = false;
    onetime = false;
    
    timer = new Timer(100, new TimerListener());
  }
  
  /***Interrompe l'esecuzione del programma per il periodo di tempo specificato in millisecondi.***/
  static public void Pause(long millis)
  {
    try
    {
      Thread.sleep(millis);
    }catch(InterruptedException e)
    {
    }
  }
  
  /***Da il via all'attivia'� del server vero e proprio.***/
  public void Start()
  {
    Initialize();
    timer.start();
  }
  
  /***Mette in esecuzione il server.***/
  public void Initialize()
  {
    System.out.println("AIServer 1.2 in esecuzione... digitare 'q' per uscire.");
    
    board = new Board();
    oldb = null;
    
    generator = new Movegen();
    generator.difficulty = diff;
    generator.tree_depth = tree_d;
    generator.test = test;   
    
    if(verbose) System.out.println("Connessione in corso...");
    
    net = new Net(addr, port, port2);
    net.SetTerm(term);
    if(net.Connect() == false)
    {
      System.err.println("Errore: Impossibile stabilire la connessione!");
      System.exit(0);
    }
    state = SERVER_STATE_CONNECTING;
  }
  
  class TimerListener implements ActionListener
  {
    public void actionPerformed(ActionEvent e)
    {
      String msg;
      int val; 
      
      if(net == null || !net.IsActive()) return;
      if(net.GetError() != Net.NET_ERROR_OK)
      {
        System.err.println("Error: "+net.GetErrorStr());
        System.exit(0);
      }
      
      if(state == SERVER_STATE_CONNECTING && verbose) System.out.println("Connessione avvenuta.");
        
      //Leggiamo il messaggio dalla connessione
      msg = net.GetMessage();
      if(msg == null) return;
      
      val = net.GetCode(msg);
      
      //Interpretiamo il messaggio
      if(val == Net.MSG_FIRST_MOVE)//La prima mossa e' concessa al programma
      {
        state = SERVER_STATE_GAME;
        if(verbose) System.out.println("Prima mossa assegnata al server AI.");
        board.SetNextMove(true);
      }
      else if(val == Net.MSG_END) //Fine partita
      {
        if(state == SERVER_STATE_CONNECTING) return;
        System.out.println("Un utente remoto ha abbandonato la partita: la partita verra' terminata!");
        
        if(onetime) System.exit(0); //Se e' attivata l'opzione onetime usciamo
        Initialize();
        return;
      }
      else if(val == Net.MSG_ERR) //La precedente mossa inviata e' stata considerata non valida
      {
        //Non dovrebbe accadere mai, aggiunto per robustezza!
        if(!oldp) board.CopyFrom(oldb);
      }
      else if(val == Net.MSG_OK) //Conferma dell'ultima mossa inviata
      {
        if(oldp) return;
        
        //Controlliamo se la mossa precedente ha portato ad una vittoria
        if(board.CheckForVictory() != 0)
        {
          board.PrintVictory(true, true);
          
          if(onetime) System.exit(0); //Se e' attivata l'opzione onetime usciamo
          Initialize();
        }      
        return;
      }
      else if(val == Net.MSG_COLOR) //Indica il colore da assumere. Non utilizzato da un AIServer
      {
        state = SERVER_STATE_GAME;
        return;
      }
      else if(val == Net.MSG_INFO) //Indica la ricezione di informazioni sul programma remoto
      {
        return;
      }
      else if(val == Net.MSG_MOVE) //Notifica di una mossa
      { 
        state = SERVER_STATE_GAME;
        String res;
        if(board.Add(Character.getNumericValue(msg.charAt(0)), Character.getNumericValue(msg.charAt(1))) == false)
        {
          System.out.println("Errore: ricevuta una mossa non valida: "+msg);
          res = net.CreateMsg(Net.MSG_ERR, 0, 0);
          if(oldp)
          {
            System.exit(0);
            return;
          }
        }
        else res = net.CreateMsg(Net.MSG_OK, 0, 0);
         
        if(!net.SendMessage(res) && !oldp) //Spediamo la risposta al messaggio solo se oldp non e' impostato a true
        {
          System.err.println("Errore: Impossibile spedire il messaggio!");
        }
        if(board.CheckForVictory() != 0)
        {
          board.PrintVictory(false, true);   
          
          if(onetime) System.exit(0); //Se e' attivata l'opzione onetime usciamo
          Initialize();
          return;
        }
      }
      else //Il messaggio arrivato non e' valido
      {
        System.err.println("Errore: ricevuto un messaggio non valido: "+msg);
      }
     
      if(state != SERVER_STATE_GAME) return;
      
      //Generiamo la mossa
      if(verbose)System.out.println("Generazione mossa in corso.");
      msg = generator.Generate(board, oldb);
      if(verbose)System.out.println("Generazione mossa effettuata: x = "+Character.toString(msg.charAt(0))+" y = "+Character.toString(msg.charAt(0)));

      if(oldb == null) oldb = new Board();
      
      oldb.CopyFrom(board); //Salviamo il vecchio stato
      msg = net.CreateMsg(Net.MSG_MOVE, Character.getNumericValue(msg.charAt(0)), Character.getNumericValue(msg.charAt(1)));
      
      //Aggiorniamo lo stato del gioco
      board.AddFast(Character.getNumericValue(msg.charAt(0)), Character.getNumericValue(msg.charAt(1)));
         
      net.SendMessage(msg);

      if(verbose)System.out.println("Invocazione del Garbage Collector.");
      System.gc();
    }
  }
}